from DBConnection import Prompt, PromptCategory, Argument, session


class Prompts:
    def add_prompt(self, prompt_name, prompt, prompt_category_name=None):
        if not prompt_category_name:
            prompt_category_name = "Default"

        prompt_category = (
            session.query(PromptCategory).filter_by(name=prompt_category_name).first()
        )
        if not prompt_category:
            prompt_category = PromptCategory(
                name=prompt_category_name,
                description=f"{prompt_category_name} category",
            )
            session.add(prompt_category)
            session.commit()

        prompt_obj = Prompt(
            name=prompt_name,
            description="",
            content=prompt,
            prompt_category=prompt_category,
        )
        session.add(prompt_obj)
        session.commit()

        # Populate prompt arguments
        prompt_args = self.get_prompt_args(prompt)
        for arg in prompt_args:
            argument = Argument(
                prompt_id=prompt_obj.id,
                name=arg,
            )
            session.add(argument)
        session.commit()

    def get_prompt(self, prompt_name, prompt_category="Default"):
        prompt = (
            session.query(Prompt)
            .filter_by(name=prompt_name)
            .join(PromptCategory)
            .filter(PromptCategory.name == prompt_category)
            .first()
        )
        if not prompt and prompt_category != "Default":
            # Prompt not found in specified category, try the default category
            prompt = (
                session.query(Prompt)
                .filter_by(name=prompt_name)
                .join(PromptCategory)
                .filter(PromptCategory.name == "Default")
                .first()
            )
        if prompt:
            return prompt.content
        return None

    def get_prompts(self):
        prompts = session.query(Prompt).all()
        return [prompt.name for prompt in prompts]

    def get_prompt_args(self, prompt_text):
        prompt_args = []
        start_index = prompt_text.find("{")
        while start_index != -1:
            end_index = prompt_text.find("}", start_index)
            if end_index != -1:
                prompt_args.append(prompt_text[start_index + 1 : end_index])
                start_index = prompt_text.find("{", end_index)
            else:
                break
        return prompt_args

    def delete_prompt(self, prompt_name):
        prompt = session.query(Prompt).filter_by(name=prompt_name).first()
        if prompt:
            # Delete associated arguments
            arguments = session.query(Argument).filter_by(prompt_id=prompt.id).all()
            for argument in arguments:
                session.delete(argument)

            session.delete(prompt)
            session.commit()

    def update_prompt(self, prompt_name, prompt, prompt_category_name=None):
        prompt_obj = session.query(Prompt).filter_by(name=prompt_name).first()
        if prompt_obj:
            if prompt_category_name:
                prompt_category = (
                    session.query(PromptCategory)
                    .filter_by(name=prompt_category_name)
                    .first()
                )
                if not prompt_category:
                    prompt_category = PromptCategory(
                        name=prompt_category_name,
                        description=f"{prompt_category_name} category",
                    )
                    session.add(prompt_category)
                    session.commit()
                prompt_obj.prompt_category = prompt_category

            prompt_obj.content = prompt
            session.commit()

            # Update prompt arguments
            prompt_args = self.get_prompt_args(prompt)
            existing_args = (
                session.query(Argument).filter_by(prompt_id=prompt_obj.id).all()
            )
            existing_arg_names = {arg.name for arg in existing_args}

            # Delete removed arguments
            for arg in existing_args:
                if arg.name not in prompt_args:
                    session.delete(arg)

            # Add new arguments
            for arg in prompt_args:
                if arg not in existing_arg_names:
                    argument = Argument(
                        prompt_id=prompt_obj.id,
                        name=arg,
                    )
                    session.add(argument)

            session.commit()

    def rename_prompt(self, prompt_name, new_prompt_name):
        prompt = session.query(Prompt).filter_by(name=prompt_name).first()
        if prompt:
            prompt.name = new_prompt_name
            session.commit()



import requests
import inspect
from chromadb.utils import embedding_functions
from chromadb.api.types import Documents, EmbeddingFunction, Embeddings
from semantic_kernel.connectors.ai.hugging_face import HuggingFaceTextEmbedding
from semantic_kernel.connectors.ai.open_ai import (
    AzureTextEmbedding,
    OpenAITextEmbedding,
)
import logging
import spacy


class LlamacppEmbeddingFunction(EmbeddingFunction):
    def __init__(self, api_host: str):
        self._api_host = api_host
        self._session = requests.Session()

    def __call__(self, texts: Documents) -> Embeddings:
        response = self._session.post(
            self._api_url, json={"content": texts, "threads": 5}
        ).json()
        if "data" in response:
            if "embedding" in response["data"]:
                return response["data"]["embedding"]
        return {}


class Embedding:
    def __init__(self, AGENT_CONFIG=None):
        self.AGENT_CONFIG = AGENT_CONFIG

    async def get_embedder(self):
        try:
            embedder = self.AGENT_CONFIG["settings"]["embedder"]
            embed, chunk_size = await self.__getattribute__(embedder)()
        except:
            embed, chunk_size = await self.default()
            logging.info("Embedder not found, using default embedder")
        return embed, chunk_size

    async def embed_text(self, text):
        embed, chunk_size = await self.get_embedder()
        return await embed(text)

    async def default(self):
        chunk_size = 128
        embed = HuggingFaceTextEmbedding(
            model_id="all-mpnet-base-v2", log=logging
        ).generate_embeddings_async
        return embed, chunk_size

    async def large_local(self):
        chunk_size = 500
        embed = HuggingFaceTextEmbedding(
            model_id="gtr-t5-large", log=logging
        ).generate_embeddings_async
        return embed, chunk_size

    async def azure(self):
        chunk_size = 1000
        embed = AzureTextEmbedding(
            deployment_name=self.AGENT_CONFIG["settings"]["AZURE_DEPLOYMENT_NAME"],
            endpoint=self.AGENT_CONFIG["settings"]["AZURE_OPENAI_ENDPOINT"],
            api_key=self.AGENT_CONFIG["settings"]["AZURE_API_KEY"],
            logger=logging,
        ).generate_embeddings_async
        return embed, chunk_size

    async def openai(self):
        chunk_size = 1000
        if "API_URI" in self.AGENT_CONFIG["settings"]:
            if self.AGENT_CONFIG["settings"]["API_URI"] != "":
                api_base = self.AGENT_CONFIG["settings"]["API_URI"]
            else:
                api_base = None
        else:
            api_base = None
        embed = OpenAITextEmbedding(
            model_id="text-embedding-ada-002",
            api_key=self.AGENT_CONFIG["settings"]["OPENAI_API_KEY"],
            endpoint=api_base,
            log=logging,
        ).generate_embeddings_async
        return embed, chunk_size

    async def google_palm(self):
        chunk_size = 1000
        embed = embedding_functions.GooglePalmEmbeddingFunction(
            api_key=self.AGENT_CONFIG["settings"]["GOOGLE_API_KEY"],
        )
        return embed, chunk_size

    async def google_vertex(self):
        chunk_size = 1000
        embed = embedding_functions.GoogleVertexEmbeddingFunction(
            api_key=self.AGENT_CONFIG["settings"]["GOOGLE_API_KEY"],
            project_id=self.AGENT_CONFIG["settings"]["GOOGLE_PROJECT_ID"],
        )
        return embed, chunk_size

    async def cohere(self):
        chunk_size = 500
        embed = embedding_functions.CohereEmbeddingFunction(
            api_key=self.AGENT_CONFIG["settings"]["COHERE_API_KEY"],
        )
        return embed, chunk_size

    async def llamacpp(self):
        chunk_size = 250
        embed = LlamacppEmbeddingFunction(
            model_name=self.AGENT_CONFIG["settings"]["EMBEDDING_URI"],
        )
        return embed, chunk_size


def get_embedding_providers():
    return [
        func
        for func, _ in inspect.getmembers(Embedding, predicate=inspect.isfunction)
        if not func.startswith("__")
    ]


def nlp(text):
    try:
        sp = spacy.load("en_core_web_sm")
    except:
        spacy.cli.download("en_core_web_sm")
        sp = spacy.load("en_core_web_sm")
    sp.max_length = 99999999999999999999999
    return sp(text)


def get_tokens(text):
    return len(nlp(text))



import re
import os
import regex
import json
import time
import logging
import asyncio
from datetime import datetime
from dotenv import load_dotenv

load_dotenv()

db_connected = True if os.getenv("DB_CONNECTED", "false").lower() == "true" else False
if db_connected:
    from db.Agent import Agent
    from db.Prompts import Prompts
    from db.Chain import Chain
    from db.History import log_interaction
else:
    from fb.Agent import Agent
    from fb.Prompts import Prompts
    from fb.Chain import Chain
    from fb.History import log_interaction

from Embedding import get_tokens
from concurrent.futures import Future
from agixtsdk import AGiXTSDK
from Websearch import Websearch

ApiClient = AGiXTSDK(
    base_uri="http://localhost:7437", api_key=os.getenv("AGIXT_API_KEY")
)
chain = Chain()
cp = Prompts()


class Interactions:
    def __init__(self, agent_name: str = ""):
        if agent_name != "":
            self.agent_name = agent_name
            self.agent = Agent(self.agent_name)
            self.agent_commands = self.agent.get_commands_string()
            self.memories = self.agent.get_memories()
            self.websearch = Websearch(agent=self.agent, memories=self.memories)
        else:
            self.agent_name = ""
            self.agent = None
            self.agent_commands = ""
            self.memories = None
        self.stop_running_event = None
        self.browsed_links = []
        self.failures = 0

    def custom_format(self, string, **kwargs):
        if isinstance(string, list):
            string = "".join(str(x) for x in string)

        def replace(match):
            key = match.group(1)
            value = kwargs.get(key, match.group(0))
            if isinstance(value, list):
                return "".join(str(x) for x in value)
            else:
                return str(value)

        pattern = r"(?<!{){([^{}\n]+)}(?!})"
        result = re.sub(pattern, replace, string)
        return result

    async def format_prompt(
        self,
        user_input: str = "",
        top_results: int = 5,
        prompt="",
        chain_name="",
        step_number=0,
        **kwargs,
    ):
        if prompt == "":
            prompt = user_input
        else:
            try:
                prompt = cp.get_prompt(
                    prompt_name=prompt,
                    prompt_category=self.agent.AGENT_CONFIG["settings"]["AI_MODEL"],
                )
            except:
                prompt = prompt
        logging.info(f"CONTEXT RESULTS: {top_results}")
        if top_results == 0:
            context = ""
        else:
            # try:
            context = await self.memories.context_agent(
                query=user_input, top_results_num=top_results
            )
            # except:
            # context = ""
        command_list = self.agent.get_commands_string()
        if chain_name != "":
            try:
                for arg, value in kwargs.items():
                    if "{STEP" in value:
                        # get the response from the step number
                        step_response = chain.get_step_response(
                            chain_name=chain_name, step_number=step_number
                        )
                        # replace the {STEPx} with the response
                        value = value.replace(f"{{STEP{step_number}}}", step_response)
                        kwargs[arg] = value
            except:
                logging.info("No args to replace.")
            if "{STEP" in prompt:
                step_response = chain.get_step_response(
                    chain_name=chain_name, step_number=step_number
                )
                prompt = prompt.replace(f"{{STEP{step_number}}}", step_response)
            if "{STEP" in user_input:
                step_response = chain.get_step_response(
                    chain_name=chain_name, step_number=step_number
                )
                user_input = user_input.replace(f"{{STEP{step_number}}}", step_response)
        try:
            working_directory = self.agent.AGENT_CONFIG["settings"]["WORKING_DIRECTORY"]
        except:
            working_directory = "./WORKSPACE"
        if "helper_agent_name" not in kwargs:
            if "helper_agent_name" in self.agent.AGENT_CONFIG["settings"]:
                helper_agent_name = self.agent.AGENT_CONFIG["settings"][
                    "helper_agent_name"
                ]
            else:
                helper_agent_name = self.agent_name
        formatted_prompt = self.custom_format(
            string=prompt,
            user_input=user_input,
            agent_name=self.agent_name,
            COMMANDS=self.agent_commands,
            context=context,
            command_list=command_list,
            date=datetime.now().strftime("%B %d, %Y %I:%M %p"),
            working_directory=working_directory,
            helper_agent_name=helper_agent_name,
            **kwargs,
        )

        tokens = get_tokens(formatted_prompt)
        logging.info(f"FORMATTED PROMPT: {formatted_prompt}")
        return formatted_prompt, prompt, tokens

    async def run(
        self,
        user_input: str = "",
        prompt: str = "",
        context_results: int = 5,
        websearch: bool = False,
        websearch_depth: int = 3,
        chain_name: str = "",
        step_number: int = 0,
        shots: int = 1,
        disable_memory: bool = False,
        conversation_name: str = "",
        browse_links: bool = False,
        **kwargs,
    ):
        shots = int(shots)
        disable_memory = True if str(disable_memory).lower() != "true" else False
        browse_links = True if str(browse_links).lower() == "true" else False
        if conversation_name != "":
            conversation_name = f"{self.agent_name} History"
        if "WEBSEARCH_TIMEOUT" in self.agent.PROVIDER_SETTINGS:
            try:
                websearch_timeout = int(
                    self.agent.PROVIDER_SETTINGS["WEBSEARCH_TIMEOUT"]
                )
            except:
                websearch_timeout = 0
        else:
            websearch_timeout = 0
        if browse_links != False:
            links = re.findall(r"(?P<url>https?://[^\s]+)", user_input)
            if links is not None and len(links) > 0:
                for link in links:
                    if link not in self.websearch.browsed_links:
                        logging.info(f"Browsing link: {link}")
                        self.websearch.browsed_links.append(link)
                        text_content, link_list = await self.memories.read_website(
                            url=link
                        )
                        if link_list is not None and len(link_list) > 0:
                            i = 0
                            for sublink in link_list:
                                if sublink[1] not in self.websearch.browsed_links:
                                    logging.info(f"Browsing link: {sublink[1]}")
                                    if i <= 10:
                                        (
                                            text_content,
                                            link_list,
                                        ) = await self.memories.read_website(
                                            url=sublink[1]
                                        )
                                        i = i + 1
        if websearch:
            if user_input == "":
                if "primary_objective" in kwargs and "task" in kwargs:
                    search_string = f"Primary Objective: {kwargs['primary_objective']}\n\nTask: {kwargs['task']}"
                else:
                    search_string = ""
            else:
                search_string = user_input
            if search_string != "":
                await self.websearch.websearch_agent(
                    user_input=search_string,
                    depth=websearch_depth,
                    timeout=websearch_timeout,
                )
        formatted_prompt, unformatted_prompt, tokens = await self.format_prompt(
            user_input=user_input,
            top_results=int(context_results),
            prompt=prompt,
            chain_name=chain_name,
            step_number=step_number,
            **kwargs,
        )
        try:
            # Workaround for non-threaded providers
            run_response = await self.agent.instruct(formatted_prompt, tokens=tokens)
            self.response = (
                run_response.result()
                if isinstance(run_response, Future)
                else run_response
            )
        except Exception as e:
            logging.info(f"Error: {e}")
            logging.info(f"PROMPT CONTENT: {formatted_prompt}")
            logging.info(f"TOKENS: {tokens}")
            self.failures += 1
            if self.failures == 5:
                self.failures == 0
                logging.info("Failed to get a response 5 times in a row.")
                return None
            logging.info(f"Retrying in 10 seconds...")
            time.sleep(10)
            if context_results > 0:
                context_results = context_results - 1
            self.response = ApiClient.prompt_agent(
                agent_name=self.agent_name,
                prompt_name=prompt,
                prompt_args={
                    "chain_name": chain_name,
                    "step_number": step_number,
                    "shots": shots,
                    "disable_memory": disable_memory,
                    "user_input": user_input,
                    "context_results": context_results,
                    "conversation_name": conversation_name,
                    **kwargs,
                },
            )

        # Handle commands if the prompt contains the {COMMANDS} placeholder
        # We handle command injection that DOESN'T allow command execution by using {command_list} in the prompt
        if "{COMMANDS}" in unformatted_prompt:
            execution_response = await self.execution_agent(
                execution_response=self.response,
                user_input=user_input,
                context_results=context_results,
                **kwargs,
            )
            return_response = ""
            if "AUTONOMOUS_EXECUTION" in self.agent.AGENT_CONFIG["settings"]:
                autonomous = (
                    True
                    if self.agent.AGENT_CONFIG["settings"]["AUTONOMOUS_EXECUTION"]
                    == True
                    else False
                )
            else:
                autonomous = False

            if autonomous == True:
                try:
                    self.response = json.loads(self.response)
                    if "response" in self.response:
                        return_response = self.response["response"]
                    if "commands" in self.response:
                        if self.response["commands"] != {}:
                            return_response += (
                                f"\n\nCommands Executed:\n{self.response['commands']}"
                            )
                    if execution_response:
                        return_response += (
                            f"\n\nCommand Execution Response:\n{execution_response}"
                        )
                except:
                    return_response = self.response
            else:
                return_response = f"{self.response}\n\n{execution_response}"
            self.response = return_response
        logging.info(f"Response: {self.response}")
        if self.response != "" and self.response != None:
            if disable_memory != True:
                try:
                    await self.memories.store_result(
                        input=user_input, result=self.response
                    )
                except:
                    pass
            if user_input != "":
                log_interaction(
                    agent_name=self.agent_name,
                    conversation_name=conversation_name,
                    role="USER",
                    message=user_input,
                )
            else:
                log_interaction(
                    agent_name=self.agent_name,
                    conversation_name=conversation_name,
                    role="USER",
                    message=formatted_prompt,
                )
            log_interaction(
                agent_name=self.agent_name,
                conversation_name=conversation_name,
                role=self.agent_name,
                message=self.response,
            )

        if shots > 1:
            responses = [self.response]
            for shot in range(shots - 1):
                shot_response = ApiClient.prompt_agent(
                    agent_name=self.agent_name,
                    prompt_name=prompt,
                    prompt_args={
                        "chain_name": chain_name,
                        "step_number": step_number,
                        "user_input": user_input,
                        "context_results": context_results,
                        "conversation_name": conversation_name,
                        **kwargs,
                    },
                )
                time.sleep(1)
                responses.append(shot_response)
            return "\n".join(
                [
                    f"Response {shot + 1}:\n{response}"
                    for shot, response in enumerate(responses)
                ]
            )
        return self.response

    # Worker Sub-Agents
    async def validation_agent(
        self, user_input, execution_response, context_results, **kwargs
    ):
        try:
            pattern = regex.compile(r"\{(?:[^{}]|(?R))*\}")
            cleaned_json = pattern.findall(execution_response)
            if len(cleaned_json) == 0:
                return {}
            if isinstance(cleaned_json, list):
                cleaned_json = cleaned_json[0]
            response = json.loads(cleaned_json)
            return response
        except:
            logging.info("INVALID JSON RESPONSE")
            logging.info(execution_response)
            logging.info("... Trying again.")
            if context_results != 0:
                context_results = context_results - 1
            else:
                context_results = 0
            execution_response = ApiClient.prompt_agent(
                agent_name=self.agent_name,
                prompt_name="JSONFormatter",
                prompt_args={
                    "user_input": user_input,
                    "context_results": context_results,
                    **kwargs,
                },
            )
            return await self.validation_agent(
                user_input=user_input,
                execution_response=execution_response,
                context_results=context_results,
                **kwargs,
            )

    def create_command_suggestion_chain(self, agent_name, command_name, command_args):
        chains = ApiClient.get_chains()
        chain_name = f"{agent_name} Command Suggestions"
        if chain_name in chains:
            step = (
                int(ApiClient.get_chain(chain_name=chain_name)["steps"][-1]["step"]) + 1
            )
        else:
            ApiClient.add_chain(chain_name=chain_name)
            step = 1
        ApiClient.add_step(
            chain_name=chain_name,
            agent_name=agent_name,
            step_number=step,
            prompt_type="Command",
            prompt={
                "command_name": command_name,
                **command_args,
            },
        )
        return f"The command has been added to a chain called '{agent_name} Command Suggestions' for you to review and execute manually."

    async def execution_agent(
        self, execution_response, user_input, context_results, **kwargs
    ):
        validated_response = await self.validation_agent(
            user_input=user_input,
            execution_response=execution_response,
            context_results=context_results,
            **kwargs,
        )
        if "commands" in validated_response:
            for command_name, command_args in validated_response["commands"].items():
                # Search for the command in the available_commands list, and if found, use the command's name attribute for execution
                if command_name is not None:
                    for available_command in self.agent.available_commands:
                        if command_name == available_command["friendly_name"]:
                            # Check if the command is a valid command in the self.avent.available_commands list
                            try:
                                if bool(self.agent.AUTONOMOUS_EXECUTION) == True:
                                    command_output = await self.agent.execute(
                                        command_name=command_name,
                                        command_args=command_args,
                                    )
                                else:
                                    command_output = (
                                        self.create_command_suggestion_chain(
                                            agent_name=self.agent_name,
                                            command_name=command_name,
                                            command_args=command_args,
                                        )
                                    )
                            except Exception as e:
                                logging.info("Command validation failed, retrying...")
                                validate_command = ApiClient.prompt_agent(
                                    agent_name=self.agent_name,
                                    prompt_name="ValidationFailed",
                                    prompt_args={
                                        "command_name": command_name,
                                        "command_args": command_args,
                                        "command_output": e,
                                        "user_input": user_input,
                                        "context_results": context_results,
                                        **kwargs,
                                    },
                                )
                                return await self.execution_agent(
                                    execution_response=validate_command,
                                    user_input=user_input,
                                    context_results=context_results,
                                    **kwargs,
                                )
                            logging.info(
                                f"Command {command_name} executed successfully with args {command_args}. Command Output: {command_output}"
                            )
                            response = f"\nExecuted Command:{command_name} with args {command_args}.\nCommand Output: {command_output}\n"
                            return response
                else:
                    if command_name == "None.":
                        return "\nNo commands were executed.\n"
                    else:
                        return f"\Command not recognized: `{command_name}`."
        else:
            return "\nNo commands were executed.\n"



from typing import List
import os
from hashlib import sha256
from Embedding import Embedding, get_tokens, nlp
from datetime import datetime
from collections import Counter
import pandas as pd
import docx2txt
import pdfplumber
from playwright.async_api import async_playwright
from semantic_kernel.connectors.memory.chroma import ChromaMemoryStore
from semantic_kernel.memory.memory_record import MemoryRecord
from chromadb.config import Settings
from bs4 import BeautifulSoup
import logging
import asyncio
import sys

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())


class Memories:
    def __init__(self, agent_name: str = "AGiXT", agent_config=None):
        self.agent_name = agent_name
        self.agent_config = agent_config
        self.chroma_client = None
        self.collection = None
        self.chunk_size = 128
        memories_dir = os.path.join(os.getcwd(), "agents", self.agent_name, "memories")
        self.chroma_client = ChromaMemoryStore(
            persist_directory=memories_dir,
            client_settings=Settings(
                chroma_db_impl="chromadb.db.duckdb.PersistentDuckDB",
                persist_directory=memories_dir,
                anonymized_telemetry=False,
            ),
        )

    async def get_embedder(self):
        embedder, chunk_size = await Embedding(
            AGENT_CONFIG=self.agent_config
        ).get_embedder()
        return embedder, chunk_size

    async def get_collection(self):
        try:
            memories_exist = await self.chroma_client.does_collection_exist_async(
                "memories"
            )
            if not memories_exist:
                await self.chroma_client.create_collection_async(
                    collection_name="memories"
                )
                memories = await self.chroma_client.get_collection_async(
                    collection_name="memories"
                )
            else:
                memories = await self.chroma_client.get_collection_async(
                    collection_name="memories"
                )
            return memories
        except Exception as e:
            raise RuntimeError(f"Unable to initialize chroma client: {e}")

    async def store_result(
        self, input: str, result: str, external_source_name: str = None
    ):
        if result:
            embedder, chunk_size = await self.get_embedder()
            collection = await self.get_collection()
            if not isinstance(result, str):
                result = str(result)
            chunks = await self.chunk_content(content=result, chunk_size=chunk_size)
            for chunk in chunks:
                record = MemoryRecord(
                    is_reference=False,
                    id=sha256(
                        (chunk + datetime.now().isoformat()).encode()
                    ).hexdigest(),
                    text=chunk,
                    timestamp=datetime.now().isoformat(),
                    description=input,
                    external_source_name=external_source_name,  # URL or File path
                    embedding=await embedder(chunk),
                    additional_metadata=chunk,
                )
                try:
                    await self.chroma_client.upsert_async(
                        collection_name="memories",
                        record=record,
                    )
                except Exception as e:
                    logging.info(f"Failed to store memory: {e}")
            self.chroma_client._client.persist()

    async def context_agent(self, query: str, top_results_num: int) -> List[str]:
        embedder, chunk_size = await self.get_embedder()
        collection = await self.get_collection()
        if collection == None:
            return []
        embed = await Embedding(AGENT_CONFIG=self.agent_config).embed_text(text=query)
        try:
            results = await self.chroma_client.get_nearest_matches_async(
                collection_name="memories",
                embedding=embed,
                limit=top_results_num,
                min_relevance_score=0.1,
            )
        except:
            return ""
        context = []
        for memory, score in results:
            context.append(memory._text)
        trimmed_context = []
        total_tokens = 0
        for item in context:
            item_tokens = get_tokens(item)
            if total_tokens + item_tokens <= chunk_size:
                trimmed_context.append(item)
                total_tokens += item_tokens
            else:
                break
        logging.info(f"Context Injected: {trimmed_context}")
        context_str = "\n".join(trimmed_context)
        response = (
            f"The user's input causes you remember these things:\n {context_str} \n\n"
        )
        return response

    def score_chunk(self, chunk: str, keywords: set):
        """Score a chunk based on the number of query keywords it contains."""
        chunk_counter = Counter(chunk.split())
        score = sum(chunk_counter[keyword] for keyword in keywords)
        return score

    async def chunk_content(
        self, content: str, chunk_size: int, overlap: int = 2
    ) -> List[str]:
        doc = nlp(content)
        sentences = list(doc.sents)
        content_chunks = []
        chunk = []
        chunk_len = 0
        keywords = [
            token.text for token in doc if token.pos_ in {"NOUN", "PROPN", "VERB"}
        ]

        for i, sentence in enumerate(sentences):
            sentence_tokens = len(sentence)
            if chunk_len + sentence_tokens > chunk_size and chunk:
                chunk_text = " ".join(token.text for token in chunk)
                content_chunks.append(
                    (self.score_chunk(chunk_text, keywords), chunk_text)
                )
                chunk = list(sentences[i - overlap : i]) if i - overlap >= 0 else []
                chunk_len = sum(len(s) for s in chunk)
            chunk.extend(sentence)
            chunk_len += sentence_tokens

        if chunk:
            chunk_text = " ".join(token.text for token in chunk)
            content_chunks.append((self.score_chunk(chunk_text, keywords), chunk_text))

        # Sort the chunks by their score in descending order before returning them
        content_chunks.sort(key=lambda x: x[0], reverse=True)
        return [chunk_text for score, chunk_text in content_chunks]

    async def read_file(self, file_path: str):
        base_path = os.path.join(os.getcwd(), "WORKSPACE")
        file_path = os.path.normpath(os.path.join(base_path, file_path))
        if not file_path.startswith(base_path):
            raise Exception("Path given not allowed")
        try:
            # If file extension is pdf, convert to text
            if file_path.endswith(".pdf"):
                with pdfplumber.open(file_path) as pdf:
                    content = "\n".join([page.extract_text() for page in pdf.pages])
            # If file extension is xls, convert to csv
            elif file_path.endswith(".xls") or file_path.endswith(".xlsx"):
                content = pd.read_excel(file_path).to_csv()
            # If file extension is doc, convert to text
            elif file_path.endswith(".doc") or file_path.endswith(".docx"):
                content = docx2txt.process(file_path)
            # TODO: If file is an image, classify it in text.
            # Otherwise just read the file
            else:
                with open(file_path, "r") as f:
                    content = f.read()
            await self.store_result(
                input=file_path, result=content, external_source_name=file_path
            )
            return True
        except:
            return False

    async def read_website(self, url):
        # try:
        async with async_playwright() as p:
            browser = await p.chromium.launch()
            context = await browser.new_context()
            page = await context.new_page()
            await page.goto(url)
            content = await page.content()

            # Scrape links and their titles
            links = await page.query_selector_all("a")
            link_list = []
            for link in links:
                title = await page.evaluate("(link) => link.textContent", link)
                href = await page.evaluate("(link) => link.href", link)
                link_list.append((title, href))

            await browser.close()
            soup = BeautifulSoup(content, "html.parser")
            text_content = soup.get_text()
            text_content = " ".join(text_content.split())
            if text_content:
                await self.store_result(
                    input=url, result=text_content, external_source_name=url
                )
            return text_content, link_list
        # except:
        #    return None, None



import os
import re
import json
import random
import requests
import logging
import asyncio
from urllib.parse import urlparse
from playwright.async_api import async_playwright
from Embedding import get_tokens
from Memories import Memories
from bs4 import BeautifulSoup
from agixtsdk import AGiXTSDK
from typing import List
from dotenv import load_dotenv

load_dotenv()
AGIXT_API_KEY = os.getenv("AGIXT_API_KEY")
db_connected = True if os.getenv("DB_CONNECTED", "false").lower() == "true" else False
if db_connected:
    from db.Agent import Agent
else:
    from fb.Agent import Agent

ApiClient = AGiXTSDK(
    base_uri="http://localhost:7437", api_key=os.getenv("AGIXT_API_KEY")
)


class Websearch:
    def __init__(
        self,
        agent: Agent,
        memories: Memories,
        **kwargs,
    ):
        self.agent = agent
        self.memories = memories
        self.agent_name = self.agent.agent_name
        try:
            self.max_tokens = self.agent.PROVIDER_SETTINGS["MAX_TOKENS"]
        except:
            self.max_tokens = 2048
        self.searx_instance_url = (
            self.agent.PROVIDER_SETTINGS["SEARXNG_INSTANCE_URL"]
            if "SEARXNG_INSTANCE_URL" in self.agent.PROVIDER_SETTINGS
            else ""
        )
        self.requirements = ["agixtsdk"]
        self.failures = []
        self.browsed_links = []
        self.tasks = []

    async def get_web_content(self, url):
        try:
            async with async_playwright() as p:
                browser = await p.chromium.launch()
                context = await browser.new_context()
                page = await context.new_page()
                await page.goto(url)
                content = await page.content()

                # Scrape links and their titles
                links = await page.query_selector_all("a")
                link_list = []
                for link in links:
                    title = await page.evaluate("(link) => link.textContent", link)
                    title = title.replace("\n", "")
                    title = title.replace("\t", "")
                    title = title.replace("  ", "")
                    href = await page.evaluate("(link) => link.href", link)
                    link_list.append((title, href))

                await browser.close()
                soup = BeautifulSoup(content, "html.parser")
                text_content = soup.get_text()
                text_content = " ".join(text_content.split())
                return text_content, link_list
        except:
            return None, None

    async def resursive_browsing(self, user_input, links):
        chunk_size = int(int(self.max_tokens) / 3)
        try:
            words = links.split()
            links = [
                word for word in words if urlparse(word).scheme in ["http", "https"]
            ]
        except:
            links = links
        if links is not None:
            for link in links:
                if "href" in link:
                    try:
                        url = link["href"]
                    except:
                        url = link
                else:
                    url = link
                url = re.sub(r"^.*?(http)", r"http", url)
                if url in self.browsed_links:
                    continue
                # Check if url is an actual url
                if url.startswith("http"):
                    logging.info(f"Scraping: {url}")
                    if url not in self.browsed_links:
                        self.browsed_links.append(url)
                        (
                            collected_data,
                            link_list,
                        ) = await self.get_web_content(url=url)
                        # Split the collected data into agent max tokens / 3 character chunks
                        if collected_data is not None:
                            if len(collected_data) > 0:
                                tokens = get_tokens(collected_data)
                                chunks = [
                                    collected_data[i : i + chunk_size]
                                    for i in range(
                                        0,
                                        int(tokens),
                                        chunk_size,
                                    )
                                ]
                                for chunk in chunks:
                                    summarized_content = ApiClient.prompt_agent(
                                        agent_name=self.agent_name,
                                        prompt_name="Summarize Web Content",
                                        prompt_args={
                                            "link": url,
                                            "chunk": chunk,
                                            "user_input": user_input,
                                            "browse_links": False,
                                            "disable_memory": True,
                                        },
                                    )
                                    if not summarized_content.startswith("None"):
                                        await self.memories.store_result(
                                            input=url,
                                            result=summarized_content,
                                            external_source_name=url,
                                        )
        if links is not None:
            for link in links:
                if "href" in link:
                    try:
                        url = link["href"]
                    except:
                        url = link
                else:
                    url = link
                url = re.sub(r"^.*?(http)", r"http", url)
                if url in self.browsed_links:
                    continue
                # Check if url is an actual url
                if url.startswith("http"):
                    logging.info(f"Scraping: {url}")
                    if url not in self.browsed_links:
                        self.browsed_links.append(url)
                        (
                            collected_data,
                            link_list,
                        ) = await self.get_web_content(url=url)
                        if link_list is not None:
                            if len(link_list) > 0:
                                if len(link_list) > 5:
                                    link_list = link_list[:3]
                                try:
                                    pick_a_link = ApiClient.prompt_agent(
                                        agent_name=self.agent_name,
                                        prompt_name="Pick-a-Link",
                                        prompt_args={
                                            "url": url,
                                            "links": link_list,
                                            "visited_links": self.browsed_links,
                                            "disable_memory": True,
                                            "browse_links": False,
                                            "user_input": user_input,
                                            "context_results": 0,
                                        },
                                    )
                                    if not pick_a_link.startswith("None"):
                                        logging.info(
                                            f"AI has decided to click: {pick_a_link}"
                                        )
                                        await self.resursive_browsing(
                                            user_input=user_input, links=pick_a_link
                                        )
                                except:
                                    logging.info(f"Issues reading {url}. Moving on...")

    async def search(self, query: str) -> List[str]:
        if self.searx_instance_url == "":
            try:  # SearXNG - List of these at https://searx.space/
                response = requests.get("https://searx.space/data/instances.json")
                data = json.loads(response.text)
                if self.failures != []:
                    for failure in self.failures:
                        if failure in data["instances"]:
                            del data["instances"][failure]
                servers = list(data["instances"].keys())
                random_index = random.randint(0, len(servers) - 1)
                self.searx_instance_url = servers[random_index]
            except:  # Select default remote server that typically works if unable to get list.
                self.searx_instance_url = "https://search.us.projectsegfau.lt"
        server = self.searx_instance_url.rstrip("/")
        endpoint = f"{server}/search"
        try:
            logging.info(f"Trying to connect to SearXNG Search at {endpoint}...")
            response = requests.get(
                endpoint,
                params={
                    "q": query,
                    "language": "en",
                    "safesearch": 1,
                    "format": "json",
                },
            )
            results = response.json()
            summaries = [
                result["title"] + " - " + result["url"] for result in results["results"]
            ]
            if len(summaries) < 1:
                self.failures.append(self.searx_instance_url)
                self.searx_instance_url = ""
                return await self.search(query=query)
            return summaries
        except:
            self.failures.append(self.searx_instance_url)
            self.searx_instance_url = ""
            # The SearXNG server is down or refusing connection, so we will use the default one.
            return await self.search(query=query)

    async def websearch_agent(
        self,
        user_input: str = "What are the latest breakthroughs in AI?",
        depth: int = 3,
        timeout: int = 0,
    ):
        results = ApiClient.prompt_agent(
            agent_name=self.agent_name,
            prompt_name="WebSearch",
            prompt_args={
                "user_input": user_input,
                "disable_memory": True,
            },
        )
        results = results.split("\n")
        if len(results) > 0:
            for result in results:
                links = []
                search_string = result.lstrip("0123456789. ")
                logging.info(f"Searching for: {search_string}")
                links = await self.search(query=search_string)
                logging.info(f"Found {len(links)} results for {search_string}")
                if len(links) > depth:
                    links = links[:depth]
                if links is not None and len(links) > 0:
                    task = asyncio.create_task(
                        self.resursive_browsing(user_input=user_input, links=links)
                    )
                    self.tasks.append(task)

            if int(timeout) == 0:
                await asyncio.gather(*self.tasks)
            else:
                logging.info(f"Web searching for {timeout} seconds... Please wait...")
                await asyncio.sleep(int(timeout))
                logging.info("Websearch tasks completed.")
        else:
            logging.info("No results found.")



import os
import json
import logging
from agixtsdk import AGiXTSDK
from Extensions import Extensions
from dotenv import load_dotenv

load_dotenv()
ApiClient = AGiXTSDK(
    base_uri="http://localhost:7437", api_key=os.getenv("AGIXT_API_KEY")
)


def create_command_suggestion_chain(agent_name, command_name, command_args):
    chain = Chain()
    chains = chain.get_chains()
    chain_name = f"{agent_name} Command Suggestions"
    if chain_name in chains:
        step = int(chain.get_chain(chain_name=chain_name)["steps"][-1]["step"]) + 1
    else:
        chain.add_chain(chain_name=chain_name)
        step = 1
    chain.add_chain_step(
        chain_name=chain_name,
        agent_name=agent_name,
        step_number=step,
        prompt_type="Command",
        prompt={
            "command_name": command_name,
            **command_args,
        },
    )
    return f"The command has been added to a chain called '{agent_name} Command Suggestions' for you to review and execute manually."


def get_chain_file_path(chain_name):
    base_path = os.path.join(os.getcwd(), "chains")
    folder_path = os.path.normpath(os.path.join(base_path, chain_name))
    file_path = os.path.normpath(os.path.join(base_path, f"{chain_name}.json"))
    if not file_path.startswith(base_path) or not folder_path.startswith(base_path):
        raise ValueError("Invalid path, chain name must not contain slashes.")
    if not os.path.exists(folder_path):
        os.mkdir(folder_path)
    return file_path


def get_chain_responses_file_path(chain_name):
    base_path = os.path.join(os.getcwd(), "chains")
    file_path = os.path.normpath(os.path.join(base_path, chain_name, "responses.json"))
    if not file_path.startswith(base_path):
        raise ValueError("Invalid path, chain name must not contain slashes.")
    return file_path


class Chain:
    def import_chain(self, chain_name: str, steps: dict):
        file_path = get_chain_file_path(chain_name=chain_name)
        steps = steps["steps"] if "steps" in steps else steps
        with open(file_path, "w") as f:
            json.dump({"chain_name": chain_name, "steps": steps}, f)
        return f"Chain '{chain_name}' imported."

    def get_chain(self, chain_name):
        try:
            file_path = get_chain_file_path(chain_name=chain_name)
            with open(file_path, "r") as f:
                chain_data = json.load(f)
            return chain_data
        except:
            return {}

    def get_chains(self):
        chains = [
            f.replace(".json", "") for f in os.listdir("chains") if f.endswith(".json")
        ]
        return chains

    def add_chain(self, chain_name):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = {"chain_name": chain_name, "steps": []}
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def rename_chain(self, chain_name, new_name):
        file_path = get_chain_file_path(chain_name=chain_name)
        new_file_path = get_chain_file_path(chain_name=new_name)
        os.rename(
            os.path.join(file_path),
            os.path.join(new_file_path),
        )
        chain_data = self.get_chain(chain_name=new_name)
        chain_data["chain_name"] = new_name
        with open(new_file_path, "w") as f:
            json.dump(chain_data, f)

    def add_chain_step(self, chain_name, step_number, agent_name, prompt_type, prompt):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        chain_data["steps"].append(
            {
                "step": step_number,
                "agent_name": agent_name,
                "prompt_type": prompt_type,
                "prompt": prompt,
            }
        )
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def update_step(self, chain_name, step_number, agent_name, prompt_type, prompt):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        for step in chain_data["steps"]:
            if step["step"] == step_number:
                step["agent_name"] = agent_name
                step["prompt_type"] = prompt_type
                step["prompt"] = prompt
                break
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def delete_step(self, chain_name, step_number):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        chain_data["steps"] = [
            step for step in chain_data["steps"] if step["step"] != step_number
        ]
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def delete_chain(self, chain_name):
        file_path = get_chain_file_path(chain_name=chain_name)
        os.remove(file_path)

    def get_step(self, chain_name, step_number):
        chain_data = self.get_chain(chain_name=chain_name)
        for step in chain_data["steps"]:
            if step["step"] == step_number:
                return step
        return None

    def get_steps(self, chain_name):
        chain_data = self.get_chain(chain_name=chain_name)
        return chain_data["steps"]

    def move_step(self, chain_name, current_step_number, new_step_number):
        file_path = get_chain_file_path(chain_name=chain_name)
        chain_data = self.get_chain(chain_name=chain_name)
        if not 1 <= new_step_number <= len(
            chain_data["steps"]
        ) or current_step_number not in [step["step"] for step in chain_data["steps"]]:
            print(f"Error: Invalid step numbers.")
            return
        moved_step = None
        for step in chain_data["steps"]:
            if step["step"] == current_step_number:
                moved_step = step
                chain_data["steps"].remove(step)
                break
        for step in chain_data["steps"]:
            if new_step_number < current_step_number:
                if new_step_number <= step["step"] < current_step_number:
                    step["step"] += 1
            else:
                if current_step_number < step["step"] <= new_step_number:
                    step["step"] -= 1
        moved_step["step"] = new_step_number
        chain_data["steps"].append(moved_step)
        chain_data["steps"] = sorted(chain_data["steps"], key=lambda x: x["step"])
        with open(file_path, "w") as f:
            json.dump(chain_data, f)

    def get_step_response(self, chain_name, step_number="all"):
        file_path = get_chain_responses_file_path(chain_name=chain_name)
        try:
            with open(file_path, "r") as f:
                responses = json.load(f)
            if step_number == "all":
                return responses
            else:
                data = responses.get(str(step_number))
                if isinstance(data, dict) and "response" in data:
                    data = data["response"]
                logging.info(f"Step {step_number} response: {data}")
                return data
        except:
            return ""

    def get_chain_responses(self, chain_name):
        file_path = get_chain_responses_file_path(chain_name=chain_name)
        try:
            with open(file_path, "r") as f:
                responses = json.load(f)
            return responses
        except:
            return {}

    def get_step_content(self, chain_name, prompt_content, user_input, agent_name):
        if isinstance(prompt_content, dict):
            new_prompt_content = {}
            for arg, value in prompt_content.items():
                if isinstance(value, str):
                    if "{user_input}" in value:
                        value = value.replace("{user_input}", user_input)
                    if "{agent_name}" in value:
                        value = value.replace("{agent_name}", agent_name)
                    if "{STEP" in value:
                        step_count = value.count("{STEP")
                        for i in range(step_count):
                            new_step_number = int(value.split("{STEP")[1].split("}")[0])
                            step_response = self.get_step_response(
                                chain_name=chain_name, step_number=new_step_number
                            )
                            if step_response:
                                resp = (
                                    step_response[0]
                                    if isinstance(step_response, list)
                                    else step_response
                                )
                                value = value.replace(
                                    f"{{STEP{new_step_number}}}", f"{resp}"
                                )
                new_prompt_content[arg] = value
            return new_prompt_content
        elif isinstance(prompt_content, str):
            new_prompt_content = prompt_content
            if "{user_input}" in prompt_content:
                new_prompt_content = new_prompt_content.replace(
                    "{user_input}", user_input
                )
            if "{agent_name}" in new_prompt_content:
                new_prompt_content = new_prompt_content.replace(
                    "{agent_name}", agent_name
                )
            if "{STEP" in prompt_content:
                step_count = prompt_content.count("{STEP")
                for i in range(step_count):
                    new_step_number = int(
                        prompt_content.split("{STEP")[1].split("}")[0]
                    )
                    step_response = self.get_step_response(
                        chain_name=chain_name, step_number=new_step_number
                    )
                    if step_response:
                        resp = (
                            step_response[0]
                            if isinstance(step_response, list)
                            else step_response
                        )
                        new_prompt_content = new_prompt_content.replace(
                            f"{{STEP{new_step_number}}}", f"{resp}"
                        )
            return new_prompt_content
        else:
            return prompt_content

    async def run_chain_step(
        self, step: dict = {}, chain_name="", user_input="", agent_override=""
    ):
        if step:
            if "prompt_type" in step:
                if agent_override != "":
                    agent_name = agent_override
                else:
                    agent_name = step["agent_name"]
                prompt_type = step["prompt_type"]
                step_number = step["step"]
                if "prompt_name" in step["prompt"]:
                    prompt_name = step["prompt"]["prompt_name"]
                else:
                    prompt_name = ""
                args = self.get_step_content(
                    chain_name=chain_name,
                    prompt_content=step["prompt"],
                    user_input=user_input,
                    agent_name=step["agent_name"],
                )
                if "conversation_name" not in args:
                    args["conversation_name"] = f"Chain Execution History: {chain_name}"
                if prompt_type == "Command":
                    return await Extensions().execute_command(
                        command_name=step["prompt"]["command_name"], command_args=args
                    )

                elif prompt_type == "Prompt":
                    result = ApiClient.prompt_agent(
                        agent_name=agent_name,
                        prompt_name=prompt_name,
                        prompt_args={
                            "chain_name": chain_name,
                            "step_number": step_number,
                            "user_input": user_input,
                            **args,
                        },
                    )
                elif prompt_type == "Chain":
                    result = ApiClient.run_chain(
                        chain_name=args["chain"],
                        user_input=args["input"],
                        agent_name=agent_name,
                        all_responses=args["all_responses"]
                        if "all_responses" in args
                        else False,
                        from_step=args["from_step"] if "from_step" in args else 1,
                    )
        if result:
            if isinstance(result, dict) and "response" in result:
                result = result["response"]
            if result == "Unable to retrieve data.":
                result = None
            return result
        else:
            return None

    async def run_chain(
        self,
        chain_name,
        user_input=None,
        all_responses=True,
        agent_override="",
        from_step=1,
    ):
        chain_data = ApiClient.get_chain(chain_name=chain_name)
        if chain_data == {}:
            return f"Chain `{chain_name}` not found."
        logging.info(f"Running chain '{chain_name}'")
        responses = {}  # Create a dictionary to hold responses.
        last_response = ""
        for step_data in chain_data["steps"]:
            if int(step_data["step"]) >= int(from_step):
                if "prompt" in step_data and "step" in step_data:
                    step = {}
                    step["agent_name"] = (
                        agent_override
                        if agent_override != ""
                        else step_data["agent_name"]
                    )
                    step["prompt_type"] = step_data["prompt_type"]
                    step["prompt"] = step_data["prompt"]
                    step["step"] = step_data["step"]
                    logging.info(
                        f"Running step {step_data['step']} with agent {step['agent_name']}."
                    )
                    try:
                        step_response = await self.run_chain_step(
                            step=step,
                            chain_name=chain_name,
                            user_input=user_input,
                            agent_override=agent_override,
                        )  # Get the response of the current step.
                    except Exception as e:
                        logging.error(e)
                        step_response = None
                    if step_response == None:
                        return f"Chain failed to complete, it failed on step {step_data['step']}. You can resume by starting the chain from the step that failed."
                    step["response"] = step_response
                    last_response = step_response
                    logging.info(f"Last response: {last_response}")
                    responses[step_data["step"]] = step  # Store the response.
                    logging.info(f"Step {step_data['step']} response: {step_response}")
                    # Write the response to the chain responses file.
                    file_path = get_chain_responses_file_path(chain_name=chain_name)
                    with open(file_path, "w") as f:
                        json.dump(responses, f)

        if all_responses:
            return responses
        else:
            # Return only the last response in the chain.
            return last_response



import os


def get_prompt_file_path(prompt_name, prompt_category="Default"):
    base_path = os.path.join(os.getcwd(), "prompts")
    base_model_path = os.path.normpath(
        os.path.join(os.getcwd(), "prompts", prompt_category)
    )
    model_prompt_file = os.path.normpath(
        os.path.join(base_model_path, f"{prompt_name}.txt")
    )
    default_prompt_file = os.path.normpath(
        os.path.join(base_path, "Default", f"{prompt_name}.txt")
    )
    if (
        not base_model_path.startswith(base_path)
        or not model_prompt_file.startswith(base_model_path)
        or not default_prompt_file.startswith(base_path)
    ):
        raise ValueError(
            "Invalid file path. Prompt name cannot contain '/', '\\' or '..' in"
        )
    if not os.path.exists(base_path):
        os.mkdir(base_path)
    if not os.path.exists(base_model_path):
        os.mkdir(base_model_path)
    prompt_file = (
        model_prompt_file if os.path.isfile(model_prompt_file) else default_prompt_file
    )
    return prompt_file


class Prompts:
    def add_prompt(self, prompt_name, prompt, prompt_category_name="Default"):
        # if prompts folder does not exist, create it
        file_path = get_prompt_file_path(
            prompt_name=prompt_name, prompt_category=prompt_category_name
        )
        # if prompt file does not exist, create it
        if not os.path.exists(file_path):
            with open(file_path, "w") as f:
                f.write(prompt)

    def get_prompt(self, prompt_name, prompt_category="Default"):
        prompt_file = get_prompt_file_path(
            prompt_name=prompt_name, prompt_category=prompt_category
        )
        with open(prompt_file, "r") as f:
            prompt = f.read()
        return prompt

    def get_prompts(self):
        # Get all files in prompts folder that end in .txt and replace .txt with empty string
        prompts = []
        # For each folder in prompts folder, get all files that end in .txt and replace .txt with empty string
        for folder in os.listdir("prompts"):
            if os.path.isdir(os.path.join("prompts", folder)):
                for file in os.listdir(os.path.join("prompts", folder)):
                    if file.endswith(".txt"):
                        prompts.append(file.replace(".txt", ""))
        return prompts

    def get_prompt_args(self, prompt_text):
        # Find anything in the file between { and } and add them to a list to return
        prompt_vars = []
        for word in prompt_text.split():
            if word.startswith("{") and word.endswith("}"):
                prompt_vars.append(word[1:-1])
        return prompt_vars

    def delete_prompt(self, prompt_name):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name)
        os.remove(prompt_file)

    def update_prompt(self, prompt_name, prompt):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name)
        with open(prompt_file, "w") as f:
            f.write(prompt)

    def rename_prompt(self, prompt_name, new_prompt_name):
        prompt_file = get_prompt_file_path(prompt_name=prompt_name)
        new_prompt_file = get_prompt_file_path(prompt_name=new_prompt_name)
        os.rename(prompt_file, new_prompt_file)



import os
import uuid
from sqlalchemy import (
    create_engine,
    Column,
    Text,
    String,
    Integer,
    ForeignKey,
    DateTime,
    Boolean,
)
from sqlalchemy.orm import sessionmaker, relationship, declarative_base
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.sql import text
from dotenv import load_dotenv

load_dotenv()


username = os.getenv("POSTGRES_USER", "postgres")
password = os.getenv("POSTGRES_PASSWORD", "postgres")
server = os.getenv("POSTGRES_SERVER", "localhost")
port = os.getenv("POSTGRES_PORT", "5432")
database_name = os.getenv("POSTGRES_DB", "postgres")
db_connected = True if os.getenv("DB_CONNECTED", "false").lower() == "true" else False
Base = declarative_base()
if db_connected:
    try:
        engine = create_engine(
            f"postgresql://{username}:{password}@{server}:{port}/{database_name}"
        )
    except Exception as e:
        print(f"Error connecting to database: {e}")
    Session = sessionmaker(bind=engine)
    session = Session()
    connection = engine.connect()
else:
    session = None


class Provider(Base):
    __tablename__ = "provider"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    provider_settings = relationship("ProviderSetting", backref="provider")


class ProviderSetting(Base):
    __tablename__ = "provider_setting"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    provider_id = Column(UUID(as_uuid=True), ForeignKey("provider.id"), nullable=False)
    name = Column(Text, nullable=False)
    value = Column(
        Text
    )  # Add the 'value' column without the 'nullable=False' constraint


class AgentProviderSetting(Base):
    __tablename__ = "agent_provider_setting"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    provider_setting_id = Column(
        UUID(as_uuid=True), ForeignKey("provider_setting.id"), nullable=False
    )
    agent_provider_id = Column(
        UUID(as_uuid=True), ForeignKey("agent_provider.id"), nullable=False
    )
    value = Column(Text, nullable=False)


class AgentProvider(Base):
    __tablename__ = "agent_provider"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    provider_id = Column(UUID(as_uuid=True), ForeignKey("provider.id"), nullable=False)
    agent_id = Column(UUID(as_uuid=True), ForeignKey("agent.id"), nullable=False)
    settings = relationship("AgentProviderSetting", backref="agent_provider")


class Agent(Base):
    __tablename__ = "agent"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    provider_id = Column(
        UUID(as_uuid=True), ForeignKey("provider.id"), nullable=True, default=None
    )
    settings = relationship("AgentSetting", backref="agent")  # One-to-many relationship


class Command(Base):
    __tablename__ = "command"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    extension_id = Column(UUID(as_uuid=True), ForeignKey("extension.id"))
    extension = relationship("Extension", backref="commands")


class AgentCommand(Base):
    __tablename__ = "agent_command"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    command_id = Column(UUID(as_uuid=True), ForeignKey("command.id"), nullable=False)
    agent_id = Column(UUID(as_uuid=True), ForeignKey("agent.id"), nullable=False)
    state = Column(Boolean, nullable=False)
    command = relationship("Command")  # Add this line to define the relationship


class Conversation(Base):
    __tablename__ = "conversation"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    agent_id = Column(UUID(as_uuid=True), ForeignKey("agent.id"), nullable=False)
    name = Column(Text, nullable=False)


class Message(Base):
    __tablename__ = "message"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    role = Column(Text, nullable=False)
    content = Column(Text, nullable=False)
    timestamp = Column(DateTime, server_default=text("now()"))
    conversation_id = Column(
        UUID(as_uuid=True), ForeignKey("conversation.id"), nullable=False
    )


class Setting(Base):
    __tablename__ = "setting"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    extension_id = Column(UUID(as_uuid=True), ForeignKey("extension.id"))
    value = Column(Text)


class AgentSetting(Base):
    __tablename__ = "agent_setting"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    agent_id = Column(UUID(as_uuid=True), ForeignKey("agent.id"), nullable=False)
    name = Column(String)
    value = Column(String)


class Chain(Base):
    __tablename__ = "chain"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    description = Column(Text, nullable=True)
    steps = relationship(
        "ChainStep",
        backref="chain",
        cascade="all, delete",  # Add the cascade option for deleting steps
        passive_deletes=True,
        foreign_keys="ChainStep.chain_id",
    )
    target_steps = relationship(
        "ChainStep", backref="target_chain", foreign_keys="ChainStep.target_chain_id"
    )


class ChainStep(Base):
    __tablename__ = "chain_step"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    chain_id = Column(
        UUID(as_uuid=True), ForeignKey("chain.id", ondelete="CASCADE"), nullable=False
    )
    agent_id = Column(UUID(as_uuid=True), ForeignKey("agent.id"), nullable=False)
    prompt_type = Column(Text)  # Add the prompt_type field
    prompt = Column(Text)  # Add the prompt field
    target_chain_id = Column(
        UUID(as_uuid=True), ForeignKey("chain.id", ondelete="SET NULL")
    )
    target_command_id = Column(
        UUID(as_uuid=True), ForeignKey("command.id", ondelete="SET NULL")
    )
    target_prompt_id = Column(
        UUID(as_uuid=True), ForeignKey("prompt.id", ondelete="SET NULL")
    )
    step_number = Column(Integer, nullable=False)
    responses = relationship(
        "ChainStepResponse", backref="chain_step", cascade="all, delete"
    )

    def add_response(self, content):
        response = ChainStepResponse(content=content, chain_step=self)
        session.add(response)
        session.commit()


class ChainStepArgument(Base):
    __tablename__ = "chain_step_argument"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    argument_id = Column(UUID(as_uuid=True), ForeignKey("argument.id"), nullable=False)
    chain_step_id = Column(
        UUID(as_uuid=True),
        ForeignKey("chain_step.id", ondelete="CASCADE"),
        nullable=False,  # Add the ondelete option
    )
    value = Column(Text, nullable=False)


class ChainStepResponse(Base):
    __tablename__ = "chain_step_response"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    chain_step_id = Column(
        UUID(as_uuid=True),
        ForeignKey("chain_step.id", ondelete="CASCADE"),
        nullable=False,  # Add the ondelete option
    )
    timestamp = Column(DateTime, server_default=text("now()"))
    content = Column(Text, nullable=False)


class Extension(Base):
    __tablename__ = "extension"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    description = Column(Text, nullable=True, default="")


class Argument(Base):
    __tablename__ = "argument"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    prompt_id = Column(UUID(as_uuid=True), ForeignKey("prompt.id"))
    command_id = Column(UUID(as_uuid=True), ForeignKey("command.id"))
    chain_id = Column(UUID(as_uuid=True), ForeignKey("chain.id"))
    name = Column(Text, nullable=False)


class PromptCategory(Base):
    __tablename__ = "prompt_category"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(Text, nullable=False)
    description = Column(Text, nullable=False)


class Prompt(Base):
    __tablename__ = "prompt"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    prompt_category_id = Column(
        UUID(as_uuid=True), ForeignKey("prompt_category.id"), nullable=False
    )
    name = Column(Text, nullable=False)
    description = Column(Text, nullable=False)
    content = Column(Text, nullable=False)

    prompt_category = relationship("PromptCategory", backref="prompts")


if __name__ == "__main__":
    Base.metadata.create_all(engine)



import os
import yaml
import json
from DBConnection import (
    session,
    Provider,
    ProviderSetting,
    Conversation,
    Message,
    Agent,
    Prompt,
    PromptCategory,
    Argument,
    Agent,
    Argument,
    Prompt,
    Extension,
    Setting,
    Command,
    Agent,
    session,
)
from Providers import get_providers, get_provider_options
from db.Agent import import_agent_config


def import_agents():
    agent_folder = "agents"
    agents = [
        f.name
        for f in os.scandir(agent_folder)
        if f.is_dir() and not f.name.startswith("__")
    ]
    existing_agents = session.query(Agent).all()
    existing_agent_names = [agent.name for agent in existing_agents]

    for agent_name in agents:
        agent = session.query(Agent).filter_by(name=agent_name).one_or_none()
        if agent:
            print(f"Updating agent: {agent_name}")
        else:
            agent = Agent(name=agent_name)
            session.add(agent)
            session.flush()  # Save the agent object to generate an ID
            existing_agent_names.append(agent_name)
            print(f"Adding agent: {agent_name}")
        import_agent_config(agent_name)
    session.commit()


def import_extensions():
    from Extensions import Extensions

    extensions_data = Extensions().get_extensions()
    extension_settings_data = Extensions().get_extension_settings()

    # Get the existing extensions and commands from the database
    existing_extensions = session.query(Extension).all()
    existing_commands = session.query(Command).all()

    # Delete commands that don't exist in the extensions data
    for command in existing_commands:
        command_exists = any(
            extension_data["extension_name"] == command.extension.name
            and any(
                cmd["friendly_name"] == command.name
                for cmd in extension_data["commands"]
            )
            for extension_data in extensions_data
        )
        if not command_exists:
            session.delete(command)

    # Add new extensions and commands, and update existing commands
    for extension_data in extensions_data:
        extension_name = extension_data["extension_name"]
        description = extension_data.get(
            "description", ""
        )  # Assign an empty string if description is missing

        # Find the existing extension or create a new one
        extension = next(
            (ext for ext in existing_extensions if ext.name == extension_name),
            None,
        )
        if extension is None:
            extension = Extension(name=extension_name, description=description)
            session.add(extension)
            session.flush()
            existing_extensions.append(extension)

        commands = extension_data["commands"]

        for command_data in commands:
            if "friendly_name" not in command_data:
                continue

            command_name = command_data["friendly_name"]

            # Find the existing command or create a new one
            command = next(
                (
                    cmd
                    for cmd in existing_commands
                    if cmd.extension_id == extension.id and cmd.name == command_name
                ),
                None,
            )
            if command is None:
                command = Command(
                    extension_id=extension.id,
                    name=command_name,
                )
                session.add(command)
                session.flush()
                existing_commands.append(command)
                print(f"Adding command: {command_name}")

            # Add command arguments
            if "command_args" in command_data:
                command_args = command_data["command_args"]
                for arg, arg_type in command_args.items():
                    if (
                        session.query(Argument)
                        .filter_by(command_id=command.id, name=arg)
                        .first()
                    ):
                        continue
                    command_arg = Argument(
                        command_id=command.id,
                        name=arg,
                    )
                    session.add(command_arg)
                    print(f"Adding argument: {arg} to command: {command_name}")

    session.commit()

    # Add extensions to the database if they don't exist
    for extension_name in extension_settings_data.keys():
        extension = session.query(Extension).filter_by(name=extension_name).first()
        if not extension:
            extension = Extension(name=extension_name)
            session.add(extension)
            session.flush()
            existing_extensions.append(extension)
            print(f"Adding extension: {extension_name}")

    session.commit()

    # Migrate extension settings
    for extension_name, settings in extension_settings_data.items():
        extension = session.query(Extension).filter_by(name=extension_name).first()
        if not extension:
            print(f"Extension '{extension_name}' not found.")
            continue

        for setting_name, setting_value in settings.items():
            setting = (
                session.query(Setting)
                .filter_by(extension_id=extension.id, name=setting_name)
                .first()
            )
            if setting:
                setting.value = setting_value
                print(
                    f"Updating setting: {setting_name} for extension: {extension_name}"
                )
            else:
                setting = Setting(
                    extension_id=extension.id,
                    name=setting_name,
                    value=setting_value,
                )
                session.add(setting)
                print(f"Adding setting: {setting_name} for extension: {extension_name}")

    session.commit()


def import_chains():
    chain_dir = os.path.abspath("chains")
    chain_files = [
        file
        for file in os.listdir(chain_dir)
        if os.path.isfile(os.path.join(chain_dir, file)) and file.endswith(".json")
    ]
    if not chain_files:
        print(f"No JSON files found in chains directory.")
        return
    from db.Chain import Chain

    chain_importer = Chain()
    for file in chain_files:
        chain_name = os.path.splitext(file)[0]
        file_path = os.path.join(chain_dir, file)

        with open(file_path, "r") as f:
            try:
                chain_data = json.load(f)
                result = chain_importer.import_chain(chain_name, chain_data)
                print(result)
            except json.JSONDecodeError as e:
                print(f"Error importing chain from '{file}': Invalid JSON format.")
            except Exception as e:
                print(f"Error importing chain from '{file}': {str(e)}")


def import_prompts():
    # Add default category if it doesn't exist
    default_category = session.query(PromptCategory).filter_by(name="Default").first()

    if not default_category:
        default_category = PromptCategory(
            name="Default", description="Default category"
        )
        session.add(default_category)
        session.commit()
        print("Adding Default prompt category")

    # Get all prompt files in the specified folder
    for root, dirs, files in os.walk("prompts"):
        for file in files:
            prompt_category = None
            if root != "prompts":
                # Use subfolder name as the prompt category
                category_name = os.path.basename(root)
                prompt_category = (
                    session.query(PromptCategory).filter_by(name=category_name).first()
                )
                if not prompt_category:
                    prompt_category = PromptCategory(
                        name=category_name, description=f"{category_name} category"
                    )
                    session.add(prompt_category)
                    session.commit()
            else:
                # Assign to "Uncategorized" category if prompt is in the root folder
                prompt_category = default_category

            # Read the prompt content from the file
            with open(os.path.join(root, file), "r") as f:
                prompt_content = f.read()

            # Check if prompt with the same name and category already exists
            prompt_name = os.path.splitext(file)[0]
            prompt = (
                session.query(Prompt)
                .filter_by(name=prompt_name, prompt_category=prompt_category)
                .first()
            )
            prompt_args = []
            for word in prompt_content.split():
                if word.startswith("{") and word.endswith("}"):
                    prompt_args.append(word[1:-1])
            if not prompt:
                # Create the prompt entry in the database
                prompt = Prompt(
                    name=prompt_name,
                    description="",
                    content=prompt_content,
                    prompt_category=prompt_category,
                )
                session.add(prompt)
                session.commit()
                print(f"Adding prompt: {prompt_name}")

            # Populate prompt arguments
            for arg in prompt_args:
                if (
                    session.query(Argument)
                    .filter_by(prompt_id=prompt.id, name=arg)
                    .first()
                ):
                    continue
                argument = Argument(
                    prompt_id=prompt.id,
                    name=arg,
                )
                session.add(argument)
                session.commit()
                print(f"Adding prompt argument: {arg} for {prompt_name}")


def import_conversations():
    agents_dir = "agents"  # Directory containing agent folders
    for agent_name in os.listdir(agents_dir):
        agent_dir = os.path.join(agents_dir, agent_name)
        history_file = os.path.join(agent_dir, "history.yaml")

        if not os.path.exists(history_file):
            continue  # Skip agent if history file doesn't exist

        # Get agent ID from the database based on agent name
        agent = session.query(Agent).filter(Agent.name == agent_name).first()
        if not agent:
            print(f"Agent '{agent_name}' not found in the database.")
            continue

        # Load conversation history from the YAML file
        with open(history_file, "r") as file:
            history = yaml.safe_load(file)

        # Check if the conversation already exists for the agent
        existing_conversation = (
            session.query(Conversation)
            .filter(
                Conversation.agent_id == agent.id,
                Conversation.name == f"{agent_name} History",
            )
            .first()
        )
        if existing_conversation:
            continue

        # Create a new conversation
        conversation = Conversation(agent_id=agent.id, name=f"{agent_name} History")
        session.add(conversation)
        session.commit()

        for conversation_data in history["interactions"]:
            # Create a new message for the conversation
            try:
                role = conversation_data["role"]
                content = conversation_data["message"]
                timestamp = conversation_data["timestamp"]
            except KeyError:
                continue
            message = Message(
                role=role,
                content=content,
                timestamp=timestamp,
                conversation_id=conversation.id,
            )
            session.add(message)
            session.commit()

        print(f"Imported `{agent_name} History` conversation for agent '{agent_name}'.")


def import_providers():
    providers = get_providers()
    existing_providers = session.query(Provider).all()
    existing_provider_names = [provider.name for provider in existing_providers]

    for provider in existing_providers:
        if provider.name not in providers:
            session.delete(provider)

    for provider_name in providers:
        provider_options = get_provider_options(provider_name)

        provider = session.query(Provider).filter_by(name=provider_name).one_or_none()

        if provider:
            print(f"Updating provider: {provider_name}")
        else:
            provider = Provider(name=provider_name)
            session.add(provider)
            existing_provider_names.append(provider_name)
            print(f"Adding provider: {provider_name}")

        for option_name, option_value in provider_options.items():
            provider_setting = (
                session.query(ProviderSetting)
                .filter_by(provider_id=provider.id, name=option_name)
                .one_or_none()
            )
            if provider_setting:
                provider_setting.value = option_value
                print(
                    f"Updating provider setting: {option_name} for provider: {provider_name}"
                )
            else:
                provider_setting = ProviderSetting(
                    provider_id=provider.id,
                    name=option_name,
                    value=option_value,
                )
                session.add(provider_setting)
                print(
                    f"Adding provider setting: {option_name} for provider: {provider_name}"
                )
    session.commit()



import os
import json
import shutil
from DBConnection import (
    Agent as AgentModel,
    AgentSetting as AgentSettingModel,
    Command,
    AgentCommand,
    ProviderSetting,
    AgentProvider,
    AgentProviderSetting,
    ChainStep,
    ChainStepArgument,
    ChainStepResponse,
    Provider as ProviderModel,
    session,
)
from Providers import Providers
from Memories import Memories
from Extensions import Extensions

DEFAULT_SETTINGS = {
    "provider": "openai",
    "embedder": "openai",
    "AI_MODEL": "gpt-3.5-turbo-16k-0613",
    "AI_TEMPERATURE": "0.7",
    "AI_TOP_P": "1",
    "MAX_TOKENS": "16000",
    "helper_agent_name": "OpenAI",
    "WEBSEARCH_TIMEOUT": 0,
    "OPENAI_API_KEY": "YOUR_OPENAI_API_KEY_HERE",
    "WAIT_BETWEEN_REQUESTS": 1,
    "WAIT_AFTER_FAILURE": 3,
    "stream": False,
    "WORKING_DIRECTORY": "./WORKSPACE",
    "WORKING_DIRECTORY_RESTRICTED": True,
    "AUTONOMOUS_EXECUTION": False,
}


def add_agent(agent_name, provider_settings=None, commands=None):
    if not agent_name:
        return {"message": "Agent name cannot be empty."}

    agent = AgentModel(name=agent_name)
    session.add(agent)
    session.commit()

    if provider_settings is None or provider_settings == "" or provider_settings == {}:
        provider_settings = DEFAULT_SETTINGS

    settings = {
        "commands": commands,
        "settings": provider_settings,
    }
    agent_setting = AgentSettingModel(
        agent_id=agent.id,
        name="config",
        value=json.dumps(settings),
    )
    session.add(agent_setting)
    session.commit()

    return {"message": f"Agent {agent_name} created."}


def delete_agent(agent_name):
    agent = session.query(AgentModel).filter_by(name=agent_name).first()
    if not agent:
        return {"message": f"Agent {agent_name} not found."}, 404

    # Delete associated chain steps
    chain_steps = session.query(ChainStep).filter_by(agent_id=agent.id).all()
    for chain_step in chain_steps:
        # Delete associated chain step arguments
        session.query(ChainStepArgument).filter_by(chain_step_id=chain_step.id).delete()
        # Delete associated chain step responses
        session.query(ChainStepResponse).filter_by(chain_step_id=chain_step.id).delete()
        session.delete(chain_step)

    # Delete associated agent commands
    agent_commands = session.query(AgentCommand).filter_by(agent_id=agent.id).all()
    for agent_command in agent_commands:
        session.delete(agent_command)

    # Delete associated agent_provider records
    agent_providers = session.query(AgentProvider).filter_by(agent_id=agent.id).all()
    for agent_provider in agent_providers:
        # Delete associated agent_provider_settings
        session.query(AgentProviderSetting).filter_by(
            agent_provider_id=agent_provider.id
        ).delete()
        session.delete(agent_provider)

    # Delete associated agent settings
    session.query(AgentSettingModel).filter_by(agent_id=agent.id).delete()

    # Delete the agent
    session.delete(agent)
    session.commit()

    return {"message": f"Agent {agent_name} deleted."}, 200


def rename_agent(agent_name, new_name):
    agent = session.query(AgentModel).filter_by(name=agent_name).first()
    if not agent:
        return {"message": f"Agent {agent_name} not found."}, 404

    agent.name = new_name
    session.commit()

    return {"message": f"Agent {agent_name} renamed to {new_name}."}, 200


def get_agents():
    agents = session.query(AgentModel).all()
    output = []

    for agent in agents:
        output.append({"name": agent.name, "status": False})

    return output


def import_agent_config(agent_name):
    config_path = f"agents/{agent_name}/config.json"

    # Load the config JSON file
    with open(config_path) as f:
        config = json.load(f)

    # Get the agent from the database
    agent = session.query(AgentModel).filter_by(name=agent_name).first()

    if not agent:
        print(f"Agent '{agent_name}' does not exist in the database.")
        return

    # Get the provider ID based on the provider name in the config
    provider_name = config["settings"]["provider"]
    provider = session.query(ProviderModel).filter_by(name=provider_name).first()

    if not provider:
        print(f"Provider '{provider_name}' does not exist in the database.")
        return

    # Update the agent's provider_id
    agent.provider_id = provider.id

    # Import agent commands
    commands = config.get("commands", {})
    for command_name, enabled in commands.items():
        if enabled:
            command = session.query(Command).filter_by(name=command_name).first()
            if command:
                agent_command = AgentCommand(
                    agent_id=agent.id, command_id=command.id, state=True
                )
                session.add(agent_command)

    # Import agent settings
    settings = config.get("settings", {})
    for setting_name, setting_value in settings.items():
        if provider.id:
            provider_setting = (
                session.query(ProviderSetting)
                .filter_by(provider_id=provider.id, name=setting_name)
                .first()
            )
            if provider_setting:
                agent_provider = (
                    session.query(AgentProvider)
                    .filter_by(provider_id=provider.id, agent_id=agent.id)
                    .first()
                )
                if not agent_provider:
                    agent_provider = AgentProvider(
                        provider_id=provider.id, agent_id=agent.id
                    )
                    session.add(agent_provider)
                    session.flush()  # Save the agent_provider object to generate an ID
                if setting_value:
                    agent_provider_setting = AgentProviderSetting(
                        provider_setting_id=provider_setting.id,
                        agent_provider_id=agent_provider.id,
                        value=setting_value,
                    )
                    session.add(agent_provider_setting)
            else:
                if setting_value:
                    agent_setting = AgentSettingModel(
                        agent_id=agent.id, name=setting_name, value=setting_value
                    )
                    session.add(agent_setting)
    session.commit()
    print(f"Agent config imported successfully for agent: {agent_name}")


class Agent:
    def __init__(self, agent_name=None):
        self.agent_name = agent_name if agent_name is not None else "AGiXT"
        self.AGENT_CONFIG = self.get_agent_config()
        self.load_config_keys()
        if "settings" not in self.AGENT_CONFIG:
            self.AGENT_CONFIG["settings"] = {}
        self.PROVIDER_SETTINGS = self.AGENT_CONFIG["settings"]
        for setting in DEFAULT_SETTINGS:
            if setting not in self.PROVIDER_SETTINGS:
                self.PROVIDER_SETTINGS[setting] = DEFAULT_SETTINGS[setting]
        self.AI_PROVIDER = self.AGENT_CONFIG["settings"]["provider"]
        self.PROVIDER = Providers(self.AI_PROVIDER, **self.PROVIDER_SETTINGS)
        self.available_commands = Extensions(
            agent_config=self.AGENT_CONFIG
        ).get_available_commands()

    def load_config_keys(self):
        config_keys = [
            "AI_MODEL",
            "AI_TEMPERATURE",
            "MAX_TOKENS",
            "AUTONOMOUS_EXECUTION",
            "embedder",
        ]
        for key in config_keys:
            if key in self.AGENT_CONFIG:
                setattr(self, key, self.AGENT_CONFIG[key])

    def get_agent_config(self):
        agent = (
            session.query(AgentModel).filter(AgentModel.name == self.agent_name).first()
        )
        if agent:
            agent_setting = (
                session.query(AgentSettingModel)
                .filter(
                    AgentSettingModel.agent_id == agent.id,
                    AgentSettingModel.name == "config",
                )
                .first()
            )
            if agent_setting:
                config = json.loads(agent_setting.value)

                # Retrieve the enabled commands for the agent
                agent_commands = (
                    session.query(AgentCommand)
                    .join(Command)
                    .filter(
                        AgentCommand.agent_id == agent.id,
                        AgentCommand.state == True,  # Only get enabled commands
                    )
                    .all()
                )
                enabled_commands = [ac.command.name for ac in agent_commands]

                # Add the enabled commands to the config
                config["enabled_commands"] = enabled_commands

                return config
        return {}

    def get_memories(self):
        return Memories(self.agent_name, self.AGENT_CONFIG)

    async def execute(self, command_name, command_args):
        return await Extensions(agent_config=self.AGENT_CONFIG).execute_command(
            command_name=command_name, command_args=command_args
        )

    async def instruct(self, prompt, tokens):
        if not prompt:
            return ""
        answer = await self.PROVIDER.instruct(prompt=prompt, tokens=tokens)
        return answer

    def get_commands_string(self):
        if len(self.available_commands) == 0:
            return None

        enabled_commands = filter(
            lambda command: command.get("enabled", True), self.available_commands
        )
        if not enabled_commands:
            return None

        friendly_names = map(
            lambda command: f"`{command['friendly_name']}` - Arguments: {command['args']}",
            enabled_commands,
        )
        command_list = "\n".join(friendly_names)
        return f"Commands Available To Complete Task:\n{command_list}\n\n"

    def update_agent_config(self, new_config, config_key):
        agent = (
            session.query(AgentModel).filter(AgentModel.name == self.agent_name).first()
        )
        if agent:
            if config_key == "commands":
                # Update AgentCommand relation
                for command_name, enabled in new_config.items():
                    command = (
                        session.query(Command).filter_by(name=command_name).first()
                    )
                    if command:
                        agent_command = (
                            session.query(AgentCommand)
                            .filter_by(agent_id=agent.id, command_id=command.id)
                            .first()
                        )
                        if agent_command:
                            agent_command.state = enabled
                        else:
                            agent_command = AgentCommand(
                                agent_id=agent.id, command_id=command.id, state=enabled
                            )
                            session.add(agent_command)
            else:
                provider = (
                    session.query(ProviderModel)
                    .filter_by(name=self.AI_PROVIDER)
                    .first()
                )
                if not provider:
                    print(
                        f"Provider '{self.AI_PROVIDER}' does not exist in the database."
                    )
                    return

                agent_provider = (
                    session.query(AgentProvider)
                    .filter_by(provider_id=provider.id, agent_id=agent.id)
                    .first()
                )
                if not agent_provider:
                    agent_provider = AgentProvider(
                        provider_id=provider.id, agent_id=agent.id
                    )
                    session.add(agent_provider)
                    session.flush()  # Save the agent_provider object to generate an ID

                if config_key in ["provider_settings", "settings"]:
                    config_dict = (
                        agent_provider.provider_settings
                        if config_key == "provider_settings"
                        else agent_provider.settings
                    )

                    for setting_name, setting_value in new_config.items():
                        setting = (
                            session.query(ProviderSetting)
                            .filter_by(provider_id=provider.id, name=setting_name)
                            .first()
                        )
                        if setting:
                            agent_provider_setting = (
                                session.query(AgentProviderSetting)
                                .filter_by(
                                    provider_setting_id=setting.id,
                                    agent_provider_id=agent_provider.id,
                                )
                                .first()
                            )
                            if agent_provider_setting:
                                agent_provider_setting.value = setting_value
                            else:
                                agent_provider_setting = AgentProviderSetting(
                                    provider_setting_id=setting.id,
                                    agent_provider_id=agent_provider.id,
                                    value=setting_value,
                                )
                                session.add(agent_provider_setting)
                else:
                    agent_setting = (
                        session.query(AgentSettingModel)
                        .filter_by(agent_id=agent.id, name=config_key)
                        .first()
                    )
                    if agent_setting:
                        agent_setting.value = new_config
                    else:
                        agent_setting = AgentSettingModel(
                            agent_id=agent.id, name=config_key, value=new_config
                        )
                        session.add(agent_setting)

            session.commit()
            return f"Agent {self.agent_name} configuration updated."
        else:
            return f"Agent {self.agent_name} not found."

    def wipe_agent_memories(self):
        memories_folder = os.path.normpath(
            os.path.join(os.getcwd(), self.agent_name, "memories")
        )
        if not memories_folder.startswith(os.getcwd()):
            raise ValueError("Invalid path, agent name must not contain slashes.")

        if os.path.exists(memories_folder):
            shutil.rmtree(memories_folder)
